@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007-2008 Christopher Lenz
+# Copyright (C) 2007-2009 Christopher Lenz
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -31,10 +31,8 @@ from inspect import getsource
 from textwrap import dedent
 import re
 import socket
-try:
-    import simplejson as json
-except ImportError:
-    import json # Python 2.6
+
+from couchdb import json
 
 __all__ = ['PreconditionFailed', 'ResourceNotFound', 'ResourceConflict',
            'ServerError', 'Server', 'Database', 'Document', 'ViewResults',
@@ -98,7 +96,7 @@ class Server(object):
 
     def __init__(self, uri=DEFAULT_BASE_URI, cache=None, timeout=None):
         """Initialize the server object.
-        
+
         :param uri: the URI of the server (for example
                     ``http://localhost:5984/``)
         :param cache: either a cache directory path (as a string) or an object
@@ -162,20 +160,44 @@ class Server(object):
         :rtype: `Database`
         :raise ResourceNotFound: if no database with that name exists
         """
-        return Database(uri(self.resource.uri, name), validate_dbname(name),
-                        http=self.resource.http)
+        db = Database(uri(self.resource.uri, name), validate_dbname(name),
+                      http=self.resource.http)
+        db.resource.head() # actually make a request to the database
+        return db
 
-    def _get_version(self):
-        resp, data = self.resource.get()
-        return data['version']
-    version = property(_get_version, doc="""\
-        The version number tuple for the CouchDB server.
+    @property
+    def config(self):
+        """The configuration of the CouchDB server.
+
+        The configuration is represented as a nested dictionary of sections and
+        options from the configuration files of the server, or the default
+        values for options that are not explicitly configured.
+
+        :type: `dict`
+        """
+        resp, data = self.resource.get('_config')
+        return data
+
+    @property
+    def version(self):
+        """The version string of the CouchDB server.
 
         Note that this results in a request being made, and can also be used
         to check for the availability of the server.
-        
-        :type: `unicode`
-        """)
+
+        :type: `unicode`"""
+        resp, data = self.resource.get()
+        return data['version']
+
+    def stats(self):
+        """Database statistics."""
+        resp, data = self.resource.get('_stats')
+        return data
+
+    def tasks(self):
+        """A list of tasks currently active on the server."""
+        resp, data = self.resource.get('_active_tasks')
+        return data
 
     def create(self, name):
         """Create a new database with the given name.
@@ -183,11 +205,32 @@ class Server(object):
         :param name: the name of the database
         :return: a `Database` object representing the created database
         :rtype: `Database`
-        :raise ResourceConflict: if a database with that name already exists
+        :raise PreconditionFailed: if a database with that name already exists
         """
         self.resource.put(validate_dbname(name))
         return self[name]
 
+    def delete(self, name):
+        """Delete the database with the specified name.
+
+        :param name: the name of the database
+        :raise ResourceNotFound: if a database with that name does not exist
+        :since: 0.6
+        """
+        del self[name]
+
+    def replicate(self, source, target, **options):
+        """Replicate changes from the source database to the target database.
+
+        :param source: URL of the source database
+        :param target: URL of the target database
+        :param options: optional replication args, e.g. continuous=True
+        """
+        data = {'source': source, 'target': target}
+        data.update(options)
+        resp, data = self.resource.post('_replicate', data)
+        return data
+
 
 class Database(object):
     """Representation of a database on a CouchDB server.
@@ -222,9 +265,9 @@ class Database(object):
     >>> doc['name'] = 'Mary Jane'
     >>> db[doc.id] = doc
 
-    The `create()` method creates a document with an auto-generated ID. If you
-    want to explicitly specify the ID, you'd use item access just as with
-    updating:
+    The `create()` method creates a document with a random ID generated by
+    CouchDB (which is not recommended). If you want to explicitly specify the
+    ID, you'd use item access just as with updating:
 
     >>> db['JohnDoe'] = {'type': 'person', 'name': 'John Doe'}
 
@@ -302,17 +345,36 @@ class Database(object):
         resp, data = self.resource.put(id, content=content)
         content.update({'_id': data['id'], '_rev': data['rev']})
 
-    def _get_name(self):
+    @property
+    def name(self):
+        """The name of the database.
+
+        Note that this may require a request to the server unless the name has
+        already been cached by the `info()` method.
+
+        :type: basestring
+        """
         if self._name is None:
-            self._name = self.info()['db_name']
+            self.info()
         return self._name
-    name = property(_get_name)
 
     def create(self, data):
-        """Create a new document in the database with a generated ID.
+        """Create a new document in the database with a random ID that is
+        generated by the server.
 
-        Any keyword arguments are used to populate the fields of the new
-        document.
+        Note that it is generally better to avoid the `create()` method and
+        instead generate document IDs on the client side. This is due to the
+        fact that the underlying HTTP ``POST`` method is not idempotent, and
+        an automatic retry due to a problem somewhere on the networking stack
+        may cause multiple documents being created in the database.
+
+        To avoid such problems you can generate a UUID on the client side.
+        Python (since version 2.5) comes with a ``uuid`` module that can be
+        used for this::
+
+            from uuid import uuid4
+            doc_id = uuid4().hex
+            db[doc_id] = {'type': 'person', 'name': 'John Doe'}
 
         :param data: the data to store in the document
         :return: the ID of the created document
@@ -321,6 +383,56 @@ class Database(object):
         resp, data = self.resource.post(content=data)
         return data['id']
 
+    def compact(self):
+        """Compact the database.
+
+        This will try to prune all revisions from the database.
+
+        :return: a boolean to indicate whether the compaction was initiated
+                 successfully
+        :rtype: `bool`
+        """
+        resp, data = self.resource.post('_compact')
+        return data['ok']
+
+    def copy(self, src, dest):
+        """Copy the given document to create a new document.
+
+        :param src: the ID of the document to copy, or a dictionary or
+                    `Document` object representing the source document.
+        :param dest: either the destination document ID as string, or a
+                     dictionary or `Document` instance of the document that
+                     should be overwritten.
+        :return: the new revision of the destination document
+        :rtype: `str`
+        :since: 0.6
+        """
+        if not isinstance(src, basestring):
+            if not isinstance(src, dict):
+                if hasattr(src, 'items'):
+                    src = src.items()
+                else:
+                    raise TypeError('expected dict or string, got %s' %
+                                    type(src))
+            src = src['_id']
+
+        if not isinstance(dest, basestring):
+            if not isinstance(dest, dict):
+                if hasattr(dest, 'items'):
+                    dest = dest.items()
+                else:
+                    raise TypeError('expected dict or string, got %s' %
+                                    type(dest))
+            if '_rev' in dest:
+                dest = '%s?%s' % (unicode_quote(dest['_id']),
+                                  unicode_urlencode({'rev': dest['_rev']}))
+            else:
+                dest = unicode_quote(dest['_id'])
+
+        resp, data = self.resource._request('COPY', src,
+                                            headers={'Destination': dest})
+        return data['rev']
+
     def delete(self, doc):
         """Delete the given document from the database.
 
@@ -343,7 +455,7 @@ class Database(object):
         ResourceConflict: ('conflict', 'Document update conflict.')
 
         >>> del server['python-tests']
-        
+
         :param doc: a dictionary or `Document` object holding the document data
         :raise ResourceConflict: if the document was updated in the database
         :since: 0.4.1
@@ -367,25 +479,46 @@ class Database(object):
         else:
             return Document(data)
 
+    def revisions(self, id, **options):
+        """Return all available revisions of the given document.
+
+        :param id: the document ID
+        :return: an iterator over Document objects, each a different revision,
+                 in reverse chronological order, if any were found
+        """
+        try:
+            resp, data = self.resource.get(id, revs=True)
+        except ResourceNotFound:
+            return
+
+        startrev = data['_revisions']['start']
+        for index, rev in enumerate(data['_revisions']['ids']):
+            options['rev'] = '%d-%s' % (startrev - index, rev)
+            revision = self.get(id, **options)
+            if revision is None:
+                return
+            yield revision
+
     def info(self):
         """Return information about the database as a dictionary.
-        
+
         The returned dictionary exactly corresponds to the JSON response to
         a ``GET`` request on the database URI.
-        
+
         :return: a dictionary of database properties
         :rtype: ``dict``
         :since: 0.4
         """
         resp, data = self.resource.get()
+        self._name = data['db_name']
         return data
 
     def delete_attachment(self, doc, filename):
         """Delete the specified attachment.
-        
-        Note that the provided `doc` is required to have a `_rev` field. Thus,
-        if the `doc` is based on a view row, the view row would need to include
-        the `_rev` field.
+
+        Note that the provided `doc` is required to have a ``_rev`` field.
+        Thus, if the `doc` is based on a view row, the view row would need to
+        include the ``_rev`` field.
 
         :param doc: the dictionary or `Document` object representing the
                     document that the attachment belongs to
@@ -397,7 +530,7 @@ class Database(object):
 
     def get_attachment(self, id_or_doc, filename, default=None):
         """Return an attachment from the specified doc id and filename.
-        
+
         :param id_or_doc: either a document ID or a dictionary or `Document`
                           object representing the document that the attachment
                           belongs to
@@ -421,9 +554,9 @@ class Database(object):
     def put_attachment(self, doc, content, filename=None, content_type=None):
         """Create or replace an attachment.
 
-        Note that the provided `doc` is required to have a `_rev` field. Thus,
+        Note that the provided `doc` is required to have a ``_rev`` field. Thus,
         if the `doc` is based on a view row, the view row would need to include
-        the `_rev` field.
+        the ``_rev`` field.
 
         :param doc: the dictionary or `Document` object representing the
                     document that the attachment should be added to
@@ -456,7 +589,7 @@ class Database(object):
     def query(self, map_fun, reduce_fun=None, language='javascript',
               wrapper=None, **options):
         """Execute an ad-hoc query (a "temp view") against the database.
-        
+
         >>> server = Server('http://localhost:5984/')
         >>> db = server.create('python-tests')
         >>> db['johndoe'] = dict(type='Person', name='John Doe')
@@ -470,18 +603,18 @@ class Database(object):
         ...     print row.key
         John Doe
         Mary Jane
-        
+
         >>> for row in db.query(map_fun, descending=True):
         ...     print row.key
         Mary Jane
         John Doe
-        
+
         >>> for row in db.query(map_fun, key='John Doe'):
         ...     print row.key
         John Doe
-        
+
         >>> del server['python-tests']
-        
+
         :param map_fun: the code of the map function
         :param reduce_fun: the code of the reduce function (optional)
         :param language: the language of the functions, to determine which view
@@ -499,7 +632,7 @@ class Database(object):
     def update(self, documents, **options):
         """Perform a bulk update or insertion of the given documents using a
         single HTTP request.
-        
+
         >>> server = Server('http://localhost:5984/')
         >>> db = server.create('python-tests')
         >>> for doc in db.update([
@@ -508,24 +641,30 @@ class Database(object):
         ...     Document(type='City', name='Gotham City')
         ... ]):
         ...     print repr(doc) #doctest: +ELLIPSIS
-        <Document '...'@'...' {'type': 'Person', 'name': 'John Doe'}>
-        <Document '...'@'...' {'type': 'Person', 'name': 'Mary Jane'}>
-        <Document '...'@'...' {'type': 'City', 'name': 'Gotham City'}>
-        
+        (True, '...', '...')
+        (True, '...', '...')
+        (True, '...', '...')
+
         >>> del server['python-tests']
-        
+
+        The return value of this method is a list containing a tuple for every
+        element in the `documents` sequence. Each tuple is of the form
+        ``(success, docid, rev_or_exc)``, where ``success`` is a boolean
+        indicating whether the update succeeded, ``docid`` is the ID of the
+        document, and ``rev_or_exc`` is either the new document revision, or
+        an exception instance (e.g. `ResourceConflict`) if the update failed.
+
         If an object in the documents list is not a dictionary, this method
         looks for an ``items()`` method that can be used to convert the object
-        to a dictionary. In this case, the returned generator will not update
-        and yield the original object, but rather yield a dictionary with
-        ``id`` and ``rev`` keys.
-        
+        to a dictionary. Effectively this means you can also use this method
+        with `schema.Document` objects.
+
         :param documents: a sequence of dictionaries or `Document` objects, or
                           objects providing a ``items()`` method that can be
                           used to convert them to a dictionary
         :return: an iterable over the resulting documents
-        :rtype: ``generator``
-        
+        :rtype: ``list``
+
         :since: version 0.2
         """
         docs = []
@@ -536,35 +675,42 @@ class Database(object):
                 docs.append(dict(doc.items()))
             else:
                 raise TypeError('expected dict, got %s' % type(doc))
+
         content = options
         content.update(docs=docs)
         resp, data = self.resource.post('_bulk_docs', content=content)
-        def _update():
-            for idx, result in enumerate(data):
-                if 'error' in result:
-                    yield result
+
+        results = []
+        for idx, result in enumerate(data):
+            if 'error' in result:
+                if result['error'] == 'conflict':
+                    exc_type = ResourceConflict
                 else:
-                    doc = documents[idx]
-                    if isinstance(doc, dict):
-                        doc.update({'_id': result['id'], '_rev': result['rev']})
-                        yield doc
-                    else:
-                        yield result
-        return _update()
+                    # XXX: Any other error types mappable to exceptions here?
+                    exc_type = ServerError
+                results.append((False, result['id'],
+                                exc_type(result['reason'])))
+            else:
+                doc = documents[idx]
+                if isinstance(doc, dict): # XXX: Is this a good idea??
+                    doc.update({'_id': result['id'], '_rev': result['rev']})
+                results.append((True, result['id'], result['rev']))
+
+        return results
 
     def view(self, name, wrapper=None, **options):
         """Execute a predefined view.
-        
+
         >>> server = Server('http://localhost:5984/')
         >>> db = server.create('python-tests')
         >>> db['gotham'] = dict(type='City', name='Gotham City')
-        
+
         >>> for row in db.view('_all_docs'):
         ...     print row.id
         gotham
-        
+
         >>> del server['python-tests']
-        
+
         :param name: the name of the view; for custom views, use the format
                      ``design_docid/viewname``, that is, the document ID of the
                      design document and the name of the view, separated by a
@@ -595,8 +741,21 @@ class Document(dict):
                                   dict([(k,v) for k,v in self.items()
                                         if k not in ('_id', '_rev')]))
 
-    id = property(lambda self: self['_id'])
-    rev = property(lambda self: self['_rev'])
+    @property
+    def id(self):
+        """The document ID.
+
+        :type: basestring
+        """
+        return self['_id']
+
+    @property
+    def rev(self):
+        """The document revision.
+
+        :type: basestring
+        """
+        return self['_rev']
 
 
 class View(object):
@@ -617,7 +776,7 @@ class View(object):
         for name, value in options.items():
             if name in ('key', 'startkey', 'endkey') \
                     or not isinstance(value, basestring):
-                value = json.dumps(value, allow_nan=False, ensure_ascii=False)
+                value = json.encode(value)
             retval[name] = value
         return retval
 
@@ -673,8 +832,7 @@ class TemporaryView(View):
         if 'keys' in options:
             options = options.copy()
             body['keys'] = options.pop('keys')
-        content = json.dumps(body, allow_nan=False,
-                             ensure_ascii=False).encode('utf-8')
+        content = json.encode(body).encode('utf-8')
         resp, data = self.resource.post(content=content, headers={
             'Content-Type': 'application/json'
         }, **self._encode_options(options))
@@ -684,10 +842,10 @@ class TemporaryView(View):
 class ViewResults(object):
     """Representation of a parameterized view (either permanent or temporary)
     and the results it produces.
-    
+
     This class allows the specification of ``key``, ``startkey``, and
     ``endkey`` options using Python slice notation.
-    
+
     >>> server = Server('http://localhost:5984/')
     >>> db = server.create('python-tests')
     >>> db['johndoe'] = dict(type='Person', name='John Doe')
@@ -701,7 +859,7 @@ class ViewResults(object):
     At this point, the view has not actually been accessed yet. It is accessed
     as soon as it is iterated over, its length is requested, or one of its
     `rows`, `total_rows`, or `offset` properties are accessed:
-    
+
     >>> len(results)
     3
 
@@ -715,11 +873,11 @@ class ViewResults(object):
     Mary Jane
     >>> people.total_rows, people.offset
     (3, 1)
-    
+
     Use plain indexed notation (without a slice) to apply the ``key`` option.
     Note that as CouchDB makes no claim that keys are unique in a view, this
     can still return multiple rows:
-    
+
     >>> list(results[['City', 'Gotham City']])
     [<Row id='gotham', key=['City', 'Gotham City'], value='Gotham City'>]
 
@@ -763,39 +921,39 @@ class ViewResults(object):
         self._total_rows = data.get('total_rows')
         self._offset = data.get('offset', 0)
 
-    def _get_rows(self):
+    @property
+    def rows(self):
+        """The list of rows returned by the view.
+
+        :type: `list`
+        """
         if self._rows is None:
             self._fetch()
         return self._rows
-    rows = property(_get_rows, doc="""\
-        The list of rows returned by the view.
-        
-        :type: `list`
-        """)
 
-    def _get_total_rows(self):
+    @property
+    def total_rows(self):
+        """The total number of rows in this view.
+
+        This value is `None` for reduce views.
+
+        :type: `int` or ``NoneType`` for reduce views
+        """
         if self._rows is None:
             self._fetch()
         return self._total_rows
-    total_rows = property(_get_total_rows, doc="""\
-        The total number of rows in this view.
-        
-        This value is `None` for reduce views.
-        
-        :type: `int` or ``NoneType`` for reduce views
-        """)
 
-    def _get_offset(self):
+    @property
+    def offset(self):
+        """The offset of the results from the first row in the view.
+
+        This value is 0 for reduce views.
+
+        :type: `int`
+        """
         if self._rows is None:
             self._fetch()
         return self._offset
-    offset = property(_get_offset, doc="""\
-        The offset of the results from the first row in the view.
-        
-        This value is 0 for reduce views.
-        
-        :type: `int`
-        """)
 
 
 class Row(dict):
@@ -808,30 +966,32 @@ class Row(dict):
         return '<%s id=%r, key=%r, value=%r>' % (type(self).__name__, self.id,
                                                  self.key, self.value)
 
-    def _get_id(self):
-        return self.get('id')
-    id = property(_get_id, doc="""\
-        The associated Document ID if it exists. Returns `None` when it
+    @property
+    def id(self):
+        """The associated Document ID if it exists. Returns `None` when it
         doesn't (reduce results).
-    """)
+        """
+        return self.get('id')
 
-    def _get_key(self):
+    @property
+    def key(self):
+        """The associated key."""
         return self['key']
-    key = property(_get_key, doc='The associated key.')
 
-    def _get_value(self):
+    @property
+    def value(self):
+        """The associated value."""
         return self['value']
-    value = property(_get_value, doc='The associated value.')
 
-    def _get_doc(self):
+    @property
+    def doc(self):
+        """The associated document for the row. This is only present when the
+        view was accessed with ``include_docs=True`` as a query parameter,
+        otherwise this property will be `None`.
+        """
         doc = self.get('doc')
         if doc:
             return Document(doc)
-    doc = property(_get_doc, doc="""\
-        The associated document for the row. This is only present when the
-        view was accessed with ``include_docs=True`` as a query parameter,
-        otherwise this property will be `None`.
-    """)
 
 
 # Internals
@@ -875,8 +1035,7 @@ class Resource(object):
         body = None
         if content is not None:
             if not isinstance(content, basestring):
-                body = json.dumps(content, allow_nan=False,
-                                  ensure_ascii=False).encode('utf-8')
+                body = json.encode(content).encode('utf-8')
                 headers.setdefault('Content-Type', 'application/json')
             else:
                 body = content
@@ -895,7 +1054,7 @@ class Resource(object):
         status_code = int(resp.status)
         if data and resp.get('content-type') == 'application/json':
             try:
-                data = json.loads(data)
+                data = json.decode(data)
             except ValueError:
                 pass
 
@@ -920,17 +1079,32 @@ def uri(base, *path, **query):
     """Assemble a uri based on a base, any number of path segments, and query
     string parameters.
 
-    >>> uri('http://example.org/', '/_all_dbs')
+    >>> uri('http://example.org', '_all_dbs')
     'http://example.org/_all_dbs'
+
+    A trailing slash on the uri base is handled gracefully:
+
+    >>> uri('http://example.org/', '_all_dbs')
+    'http://example.org/_all_dbs'
+
+    And multiple positional arguments become path parts:
+
+    >>> uri('http://example.org/', 'foo', 'bar')
+    'http://example.org/foo/bar'
+
+    All slashes within a path part are escaped:
+
+    >>> uri('http://example.org/', 'foo/bar')
+    'http://example.org/foo%2Fbar'
+    >>> uri('http://example.org/', 'foo', '/bar/')
+    'http://example.org/foo/%2Fbar%2F'
     """
     if base and base.endswith('/'):
         base = base[:-1]
     retval = [base]
 
     # build the path
-    path = '/'.join([''] +
-                    [unicode_quote(s.strip('/')) for s in path
-                     if s is not None])
+    path = '/'.join([''] + [unicode_quote(s) for s in path if s is not None])
     if path:
         retval.append(path)
 
@@ -968,7 +1142,7 @@ def unicode_urlencode(data):
     return urlencode(params)
 
 
-VALID_DB_NAME = re.compile(r'^[a-z0-9_$()+-/]+$')
+VALID_DB_NAME = re.compile(r'^[a-z][a-z0-9_$()+-/]*$')
 def validate_dbname(name):
     if not VALID_DB_NAME.match(name):
         raise ValueError('Invalid database name')
